"""
Energy and Work Dimension Module.

This module defines units representing the capacity to do work or produce heat.
It covers mechanical work, electrical energy, thermodynamics, food energy, 
and quantum physics (e.g., Joules, Calories, Watt-hours, Electronvolts).
The absolute base unit for this dimension is the Joule (J).
"""

from ..core.base import BaseUnit
from ..core import axioms as axiom
from ..core import constants as const

from .force import Newton, PoundForce
from .length import Meter, Foot


@axiom.bound(min_val=0, msg="Absolute energy (scalar) cannot have a negative value.")
class EnergyUnit(BaseUnit):
    """The primary parent class for Energy and Work dimensions. The base unit is Joule (J)."""
    dimension = "energy"


# =========================================================================
# 1. SI / METRIC UNITS (Mechanical Energy)
# =========================================================================
# Joule: 1 Newton * 1 Meter
@axiom.derive(mul=[Newton, Meter])
class Joule(EnergyUnit):
    symbol = "J"
    aliases = ["joule", "joules"]

@axiom.derive(mul=[1e3, Joule])
class Kilojoule(EnergyUnit):
    symbol = "kJ"
    aliases = ["kilojoule", "kilojoules"]

@axiom.derive(mul=[1e6, Joule])
class Megajoule(EnergyUnit):
    symbol = "MJ"
    aliases = ["megajoule", "megajoules"]

@axiom.derive(mul=[1e9, Joule])
class Gigajoule(EnergyUnit):
    symbol = "GJ"
    aliases = ["gigajoule", "gigajoules"]


# =========================================================================
# 2. THERMODYNAMICS & FOOD (Calories)
# =========================================================================
@axiom.derive(mul=[const.CALORIE_TO_JOULE, Joule])
class Calorie(EnergyUnit):
    symbol = "cal"
    aliases = ["calorie", "calories", "gram calorie"]

@axiom.derive(mul=[const.CALORIE_TO_JOULE * 1000.0, Joule])
class Kilocalorie(EnergyUnit):
    symbol = "kcal"
    aliases = ["kilocalorie", "kilocalories", "food calorie", "Cal"]


# =========================================================================
# 3. ELECTRICAL ENERGY
# =========================================================================
# Watt-hour is defined as 1 Watt over 1 Hour
@axiom.derive(mul=[const.HOUR_TO_SECOND, Joule])
class WattHour(EnergyUnit):
    symbol = "Wh"
    aliases = ["watt-hour", "watthour", "watt hours"]

@axiom.derive(mul=[const.HOUR_TO_SECOND * 1000.0, Joule])
class KilowattHour(EnergyUnit):
    symbol = "kWh"
    aliases = ["kilowatt-hour", "kilowatt hour", "kilowatt hours", "kwh"]

@axiom.derive(mul=[const.HOUR_TO_SECOND * 1e6, Joule])
class MegawattHour(EnergyUnit):
    symbol = "MWh"
    aliases = ["megawatt-hour", "megawatt hour", "megawatt hours"]


# =========================================================================
# 4. IMPERIAL / US CUSTOMARY & HVAC
# =========================================================================
# British Thermal Unit (International Table)
@axiom.derive(mul=[const.BTU_TO_JOULE, Joule])
class BTU(EnergyUnit):
    symbol = "BTU"
    aliases = ["btu", "british thermal unit"]

# Foot-pound force: Imperial mechanical work (1 lbf applied over 1 foot)
@axiom.derive(mul=[PoundForce, Foot])
class FootPound(EnergyUnit):
    symbol = "ft-lbf"
    aliases = ["foot-pound", "foot-pounds", "ft·lbf", "ftlbf"]


# =========================================================================
# 5. QUANTUM / ATOMIC PHYSICS
# =========================================================================
# Electronvolt: Kinetic energy gained by a single electron
@axiom.derive(mul=[const.ELECTRON_VOLT_TO_JOULE, Joule])
class Electronvolt(EnergyUnit):
    symbol = "eV"
    aliases = ["electronvolt", "electronvolts", "ev"]


"""
Mass Dimension Module.

This module provides units for quantifying the amount of matter in an object.
It supports everyday metric and imperial weights, jewelry standards (Carats), 
microscopic scales (Daltons, Planck mass), and cosmic scales (Solar Mass, Earth Mass).
The absolute base unit for this dimension is the Kilogram (kg).
"""

from ..core.base import BaseUnit
from ..core import axioms as axiom
from ..core import constants as const # <-- The SSOT

@axiom.bound(min_val=0, msg="Mass cannot be a negative value.")
class MassUnit(BaseUnit):
    """
    The primary parent class for the Mass dimension.
    The base unit is Kilogram (kg).
    """
    dimension = "mass"

# =========================================================================
# 1. METRIC UNITS (SI) 
# (Metric is intrinsically base-10, so magic numbers are acceptable here)
# =========================================================================
class Kilogram(MassUnit):
    symbol = "kg"
    aliases = ["kilogram", "kilograms"]
    base_multiplier = 1.0

class Gram(MassUnit):
    symbol = "g"
    aliases = ["gram", "grams"]
    base_multiplier = 1e-3

class Milligram(MassUnit):
    symbol = "mg"
    aliases = ["milligram", "milligrams"]
    base_multiplier = 1e-6

class MetricTon(MassUnit):
    symbol = "t"
    aliases = ["ton", "tons", "metricton"]
    base_multiplier = 1000.0

class Quintal(MassUnit):
    symbol = "quintal"
    base_multiplier = 100.0

class Ons(MassUnit):
    symbol = "ons"
    aliases = ["ons-nl"]
    base_multiplier = 0.1


# =========================================================================
# 2. IMPERIAL & US CUSTOMARY UNITS
# =========================================================================
class Pound(MassUnit):
    symbol = "lb"
    aliases = ["pound", "pounds"]
    base_multiplier = const.POUND_TO_KG

class Ounce(MassUnit):
    symbol = "oz"
    aliases = ["ounce", "ounces"]
    base_multiplier = const.OUNCE_TO_KG

class TroyOunce(MassUnit):
    symbol = "oz_t"
    aliases = ["ozt", "troy ounce", "troy ounces"]
    base_multiplier = const.TROY_OUNCE_TO_KG

class Stone(MassUnit):
    symbol = "st"
    aliases = ["stone", "stones"]
    base_multiplier = const.STONE_TO_KG

class Slug(MassUnit):
    symbol = "slug"
    base_multiplier = const.SLUG_TO_KG

class Dram(MassUnit):
    symbol = "dr"
    aliases = ["dram", "drams"]
    base_multiplier = const.POUND_TO_KG / 256.0

class ShortTon(MassUnit):
    symbol = "shortton"
    aliases = ["short ton"]
    base_multiplier = const.SHORT_TON_TO_KG

class LongTon(MassUnit):
    symbol = "longton"
    aliases = ["long ton"]
    base_multiplier = const.LONG_TON_TO_KG


# =========================================================================
# 3. SMALLER / JEWELRY UNITS
# =========================================================================
class Carat(MassUnit):
    symbol = "carat"
    aliases = ["carats"]
    base_multiplier = const.CARAT_TO_KG

class Grain(MassUnit):
    symbol = "grain"
    aliases = ["grains"]
    base_multiplier = const.GRAIN_TO_KG


# =========================================================================
# 4. ASTRONOMICAL MASS UNITS
# =========================================================================
class SolarMass(MassUnit):
    symbol = "M☉"
    aliases = ["solarmass", "solar-mass"]
    base_multiplier = const.SOLAR_MASS_KG

class EarthMass(MassUnit):
    symbol = "M⊕"
    aliases = ["earthmass", "earth-mass"]
    base_multiplier = const.EARTH_MASS_KG

class LunarMass(MassUnit):
    symbol = "M☾"
    aliases = ["lunarmass", "lunar-mass"]
    base_multiplier = const.LUNAR_MASS_KG

class JupiterMass(MassUnit):
    symbol = "Mj"
    aliases = ["jupitermass", "jupiter-mass", "M♃"]
    base_multiplier = const.JUPITER_MASS_KG

class SaturnMass(MassUnit):
    symbol = "Msat"
    aliases = ["saturnmass", "saturn-mass", "M♄"]
    base_multiplier = const.SATURN_MASS_KG

class UranusMass(MassUnit):
    symbol = "Mura"
    aliases = ["uranusmass", "uranus-mass"]
    base_multiplier = const.URANUS_MASS_KG

class NeptuneMass(MassUnit):
    symbol = "Mnep"
    aliases = ["neptunemass", "neptune-mass"]
    base_multiplier = const.NEPTUNE_MASS_KG

class VenusMass(MassUnit):
    symbol = "Mven"
    aliases = ["venusmass", "venus-mass"]
    base_multiplier = const.VENUS_MASS_KG

class MarsMass(MassUnit):
    symbol = "Mmars"
    aliases = ["marsmass", "mars-mass"]
    base_multiplier = const.MARS_MASS_KG

class MercuryMass(MassUnit):
    symbol = "Mmer"
    aliases = ["mercurymass", "mercury-mass"]
    base_multiplier = const.MERCURY_MASS_KG

class PlutoMass(MassUnit):
    symbol = "Mplu"
    aliases = ["plutomass", "pluto-mass"]
    base_multiplier = const.PLUTO_MASS_KG

class CeresMass(MassUnit):
    symbol = "Mcer"
    aliases = ["ceresmass", "ceres-mass"]
    base_multiplier = const.CERES_MASS_KG


# =========================================================================
# 5. ATOMIC / QUANTUM UNITS
# =========================================================================
class AtomicMassUnit(MassUnit):
    symbol = "amu"
    aliases = ["u", "Da", "dalton", "atomicmassunit", "atomic mass unit", "atomic-mass-unit"]
    base_multiplier = const.ATOMIC_MASS_UNIT_KG

class PlanckMass(MassUnit):
    symbol = "m_p"
    aliases = ["planckmass", "planck-mass", "planck mass"]
    base_multiplier = const.PLANCK_MASS_KG

class ElectronVoltPerCSquared(MassUnit):
    symbol = "eV/c^2"
    aliases = [
        "ev/c^2", "ev/c2", "ev/c²", "eV/c2",
        "electronvoltpercsquared", "electron-volt-per-c-squared",
        "electron volt per c squared", "electron volt/c^2", "electron volt/c2"
    ]
    base_multiplier = const.ELECTRON_VOLT_C2_KG


# =========================================================================
# 6. OBSOLETE / REGIONAL UNITS
# =========================================================================
class BaleCotton(MassUnit):
    symbol = "bale-cotton"
    aliases = ["bale", "bale_cotton"]
    base_multiplier = 217.72

class BaleWool(MassUnit):
    symbol = "bale-wool"
    aliases = ["bale_wool"]
    base_multiplier = 204.0

class BaleUK(MassUnit):
    symbol = "bale-uk"
    aliases = ["bale_uk"]
    base_multiplier = 226.8

class BaleAus(MassUnit):
    symbol = "bale-aus"
    aliases = ["bale_aus"]
    base_multiplier = 204.0

class MarkDE(MassUnit):
    symbol = "mark-de"
    aliases = ["mark", "mark_de"]
    base_multiplier = 0.25

class MarkNO(MassUnit):
    symbol = "mark-no"
    aliases = ["mark_no"]
    base_multiplier = 0.213

class ArrobaES(MassUnit):
    symbol = "arroba-es"
    aliases = ["arroba", "arroba_es"]
    base_multiplier = 11.5

class ArrobaPT(MassUnit):
    symbol = "arroba-pt"
    aliases = ["arrobas", "arroba_pt"]
    base_multiplier = 15.0


"""
Force Dimension Module.

This module implements units of physical force, bridging absolute systems 
(like Newtons and Dynes) with gravitational systems (like Kilogram-force 
and Pound-force) based on standard Earth gravity.
The absolute base unit for this dimension is the Newton (N).
"""

from ..core.base import BaseUnit
from ..core import axioms as axiom
from ..core import constants as const

from .mass import Kilogram, Gram, Pound, Ounce, MetricTon
from .length import Meter, Centimeter, Foot
from .time import Second


@axiom.bound(min_val=0, msg="The scalar magnitude of force cannot be negative.")
class ForceUnit(BaseUnit):
    """The primary parent class for the Force dimension. The base unit is Newton (N)."""
    dimension = "force"


# =========================================================================
# 1. SI / METRIC UNITS (Absolute)
# =========================================================================
# Newton: 1 kg * m / s²
@axiom.derive(mul=[Kilogram, Meter], div=[Second, Second])
class Newton(ForceUnit):
    symbol = "N"
    aliases = ["newton", "newtons"]

@axiom.derive(mul=[1000, Newton])
class Kilonewton(ForceUnit):
    symbol = "kN"
    aliases = ["kilonewton", "kilonewtons"]

@axiom.derive(mul=[1e6, Newton])
class Meganewton(ForceUnit):
    symbol = "MN"
    aliases = ["meganewton", "meganewtons"]

# Dyne (CGS System): 1 g * cm / s² (0.00001 N)
@axiom.derive(mul=[Gram, Centimeter], div=[Second, Second])
class Dyne(ForceUnit):
    symbol = "dyn"
    aliases = ["dyne", "dynes"]


# =========================================================================
# 2. GRAVITATIONAL METRIC UNITS
# =========================================================================
# Kilogram-force (Kilopond): 1 kg * STANDARD_GRAVITY
@axiom.derive(mul=[Kilogram, const.STANDARD_GRAVITY, Meter], div=[Second, Second])
class KilogramForce(ForceUnit):
    symbol = "kgf"
    aliases = ["kilogram-force", "kilopond", "kp"]

@axiom.derive(mul=[Gram, const.STANDARD_GRAVITY, Meter], div=[Second, Second])
class GramForce(ForceUnit):
    symbol = "gf"
    aliases = ["gram-force", "pond", "p"]

@axiom.derive(mul=[MetricTon, const.STANDARD_GRAVITY, Meter], div=[Second, Second])
class TonneForce(ForceUnit):
    symbol = "tf"
    aliases = ["tonne-force", "ton-force", "metric ton-force"]


# =========================================================================
# 3. US CUSTOMARY / IMPERIAL UNITS
# =========================================================================
# Pound-force: 1 lb * standard Earth gravity
@axiom.derive(mul=[Pound, const.STANDARD_GRAVITY, Meter], div=[Second, Second])
class PoundForce(ForceUnit):
    symbol = "lbf"
    aliases = ["pound-force", "pounds-force"]

# Ounce-force: 1 oz * standard Earth gravity
@axiom.derive(mul=[Ounce, const.STANDARD_GRAVITY, Meter], div=[Second, Second])
class OunceForce(ForceUnit):
    symbol = "ozf"
    aliases = ["ounce-force", "ounces-force"]

# Poundal (Absolute Imperial): 1 lb * 1 ft / s²
@axiom.derive(mul=[Pound, Foot], div=[Second, Second])
class Poundal(ForceUnit):
    symbol = "pdl"
    aliases = ["poundal", "poundals"]

# Ton-force (Short Ton / US): 2000 lb * standard Earth gravity
@axiom.derive(mul=[2000, Pound, const.STANDARD_GRAVITY, Meter], div=[Second, Second])
class ShortTonForce(ForceUnit):
    symbol = "short tonf"
    aliases = ["tonf", "us ton-force", "short ton-force"]


"""
Digital Data and Information Dimension Module.

This module provides units for measuring digital storage and information entropy.
It safely handles the mathematical distinctions between base-10 SI decimal bytes 
(e.g., Kilobyte, Megabyte) and base-2 IEC binary bytes (e.g., Kibibyte, Mebibyte).
The absolute base unit for this dimension is the Byte (B).
"""

from ..core.base import BaseUnit
from ..core import axioms as axiom
from ..core import constants as const


@axiom.bound(min_val=0, msg="Data size cannot be negative!")
class DataUnit(BaseUnit):
    """
    The primary parent class for all digital information and data storage units.
    The base unit is Byte (B).
    """
    dimension = "data"


# =========================================================================
# 1. BITS & NIBBLE
# =========================================================================
class Bit(DataUnit):
    symbol = "b"
    aliases = ["bit", "bits", "bps"]
    base_multiplier = const.BYTES_PER_BIT

class Nibble(DataUnit):
    symbol = "nibble"
    aliases = ["nibbles"]
    base_multiplier = const.BYTES_PER_NIBBLE

class Byte(DataUnit):
    """The fundamental base unit for the data dimension."""
    symbol = "B"
    aliases = ["byte", "bytes", "Bps"]
    base_multiplier = 1.0


# =========================================================================
# 2. METRIC / SI BYTES (Base 10 / 1000)
# (Scientific notation '1eX' is mathematically standard and clean here)
# =========================================================================
class Kilobyte(DataUnit):
    symbol = "KB"
    aliases = ["kilobyte", "kilobytes", "kbyte", "KBps"]
    base_multiplier = 1e3

class Megabyte(DataUnit):
    symbol = "MB"
    aliases = ["megabyte", "megabytes", "mbyte", "MBps"]
    base_multiplier = 1e6

class Gigabyte(DataUnit):
    symbol = "GB"
    aliases = ["gigabyte", "gigabytes", "gbyte", "GBps"]
    base_multiplier = 1e9

class Terabyte(DataUnit):
    symbol = "TB"
    aliases = ["terabyte", "terabytes", "tbyte", "TBps"]
    base_multiplier = 1e12

class Petabyte(DataUnit):
    symbol = "PB"
    aliases = ["petabyte", "petabytes", "pbyte", "PBps"]
    base_multiplier = 1e15


# =========================================================================
# 3. BINARY / IEC BYTES (Base 2 / 1024)
# =========================================================================
class Kibibyte(DataUnit):
    symbol = "KiB"
    aliases = ["kibibyte", "kibibytes", "kib"]
    base_multiplier = const.IEC_BASE

class Mebibyte(DataUnit):
    symbol = "MiB"
    aliases = ["mebibyte", "mebibytes", "mib"]
    base_multiplier = const.IEC_BASE ** 2

class Gibibyte(DataUnit):
    symbol = "GiB"
    aliases = ["gibibyte", "gibibytes", "gib"]
    base_multiplier = const.IEC_BASE ** 3

class Tebibyte(DataUnit):
    symbol = "TiB"
    aliases = ["tebibyte", "tebibytes", "tib"]
    base_multiplier = const.IEC_BASE ** 4

class Pebibyte(DataUnit):
    symbol = "PiB"
    aliases = ["pebibyte", "pebibytes", "pib"]
    base_multiplier = const.IEC_BASE ** 5


# =========================================================================
# 4. METRIC BITS
# =========================================================================
class Kilobit(DataUnit):
    symbol = "Kb"
    aliases = ["kilobit", "kilobits", "kbit", "kbps"]
    # 1000 bits converted to Bytes
    base_multiplier = 1e3 * const.BYTES_PER_BIT

class Megabit(DataUnit):
    symbol = "Mb"
    aliases = ["megabit", "megabits", "mbit", "mbps"]
    base_multiplier = 1e6 * const.BYTES_PER_BIT

class Gigabit(DataUnit):
    symbol = "Gb"
    aliases = ["gigabit", "gigabits", "gbit", "gbps"]
    base_multiplier = 1e9 * const.BYTES_PER_BIT

class Terabit(DataUnit):
    symbol = "Tb"
    aliases = ["terabit", "terabits", "tbit", "tbps"]
    base_multiplier = 1e12 * const.BYTES_PER_BIT


from chisa import convert, axiom
from chisa.units.time import Hour, Minute, TimeUnit, Second, Millennium, Year
from chisa.units.force import Newton
from chisa.exceptions import AxiomViolationError

res = Hour(-1.3) + Minute(20)
print(res.flex())

res = convert(1.3, 'hour').to('kilometers')
print(res)

@axiom.require(waktu='time')
def konversi_ke_detik(waktu: TimeUnit) -> Second:
    return waktu.to(Second)

print(konversi_ke_detik(Millennium()))

@axiom.require(tahun=Year)
def konversi_usia_ke_detik(tahun: Year) -> Second:
    return Year.to(Second)

print(konversi_usia_ke_detik(Newton(20)))



"""
Power Dimension Module.

This module measures the rate of energy transfer or work done over time.
It includes standard electrical/mechanical power (Watts), automotive mechanics 
(Horsepower), and standard HVAC/refrigeration capacities (BTU/h, Tons of Refrigeration).
The absolute base unit for this dimension is the Watt (W).
"""

from ..core.base import BaseUnit
from ..core import axioms as axiom

from .energy import Joule, FootPound, BTU
from .time import Second, Hour
from .force import KilogramForce
from .length import Meter


@axiom.bound(min_val=0, msg="Power magnitude cannot be negative.")
class PowerUnit(BaseUnit):
    """The primary parent class for the Power dimension. The base unit is Watt (W)."""
    dimension = "power"


# =========================================================================
# 1. SI / METRIC UNITS (Electrical & Mechanical)
# =========================================================================
# Watt: 1 Joule per Second
@axiom.derive(mul=[Joule], div=[Second])
class Watt(PowerUnit):
    symbol = "W"
    aliases = ["watt", "watts"]

@axiom.derive(mul=[1e3, Watt])
class Kilowatt(PowerUnit):
    symbol = "kW"
    aliases = ["kilowatt", "kilowatts"]

@axiom.derive(mul=[1e6, Watt])
class Megawatt(PowerUnit):
    symbol = "MW"
    aliases = ["megawatt", "megawatts"]

@axiom.derive(mul=[1e9, Watt])
class Gigawatt(PowerUnit):
    symbol = "GW"
    aliases = ["gigawatt", "gigawatts"]


# =========================================================================
# 2. HORSEPOWER
# =========================================================================
# Mechanical Horsepower (Imperial): Defined as 550 foot-pounds per second.
# Chisa dynamically synthesizes this to ~745.6998 Watts.
@axiom.derive(mul=[550, FootPound], div=[Second])
class Horsepower(PowerUnit):
    symbol = "hp"
    aliases = ["horsepower", "mechanical horsepower", "imperial horsepower"]

# Metric Horsepower (PS / Pferdestärke): Defined as 75 kgf·m per second.
# Chisa dynamically synthesizes this to ~735.49875 Watts.
@axiom.derive(mul=[75, KilogramForce, Meter], div=[Second])
class MetricHorsepower(PowerUnit):
    symbol = "PS"
    aliases = ["ps", "metric horsepower", "ch", "pk"]


# =========================================================================
# 3. HVAC / REFRIGERATION
# =========================================================================
# BTU per hour: Standard for Air Conditioner cooling capacity.
@axiom.derive(mul=[BTU], div=[Hour])
class BTUPerHour(PowerUnit):
    symbol = "BTU/h"
    aliases = ["btu/h", "btuh", "btu per hour"]

# Ton of Refrigeration (Freezing capacity of 1 short ton of ice in 24 hours).
# Defined as exactly 12,000 BTU/h.
@axiom.derive(mul=[12000, BTUPerHour])
class TonOfRefrigeration(PowerUnit):
    symbol = "TR"
    aliases = ["tr", "ton of refrigeration", "tons of refrigeration"]


"""
Chisa is a logic-driven Dimensional Algebra and Physics Modeling Framework for Python.

Chisa provides a robust, developer-friendly ecosystem for handling physical and 
digital unit conversions, mathematical operations across different units, and 
strict dimensional validation. It is designed to prevent dimensional mismatch 
errors in scientific computing, engineering applications, and everyday scripting.

Scope and Capabilities:
    Chisa covers a wide array of physical and digital dimensions, including:
    Area, Data, Energy, Force, Length, Mass, Power, Pressure, Speed, 
    Temperature, Time, and Volume.

Core Paradigms:

    1. The Fluent API (Quick Conversions)
       A chainable, readable interface for straightforward conversions and string formatting.
       >>> from chisa import convert
       >>> result = convert(1.5, 'kg').to('g').use(format='verbose').resolve()
       >>> print(result) # "1.5 kg = 1500 g"

    2. Explicit OOP (Dimensional Math & State)
       Treat units as first-class objects. Chisa automatically handles base-value 
       normalization, allowing seamless math and comparisons between different units 
       of the same dimension.
       >>> from chisa.units.length import Meter, Centimeter
       >>> total = Meter(1) + Centimeter(50)
       >>> print(total) # <Meter: 1.5 m>

    3. The Axiom Engine (Strict Physics Modeling)
       Developers can extend Chisa by defining custom units or entirely new dimensions 
       by inheriting from `BaseUnit`. The `@axiom` decorators act as a physics rule engine
       to model real-world physical laws:
       
       >>> from chisa import axiom
       >>> from chisa.units.energy import Joule
       >>> from chisa.units.time import Second
       >>> from chisa.units.power import PowerUnit
       >>> 
       >>> # Synthesize a new unit based on physical laws (P = W / t)
       >>> @axiom.derive(mul=[Joule], div=[Second])
       >>> class Watt(PowerUnit):
       ...     symbol = "W"
       ...     aliases = ["watt", "watts"]

       - `@axiom.derive`: Synthesizes base multipliers dynamically.
       - `@axiom.bound`: Enforces physical limits (e.g., Absolute Zero for Kelvin).
       - `@axiom.scale` / `.shift`: Models context-dependent unit scaling (e.g., Mach).
       - `@axiom.require`: Acts as a guardrail for custom physics functions.

    4. Vectorized Computations (NumPy Integration)
       Seamlessly supports multidimensional NumPy arrays for high-performance 
       scientific computing without losing dimensional integrity.
       >>> import numpy as np
       >>> speeds = convert(np.array([1, 2, 3]), 'mach').to('m/s').resolve()
"""

from chisa.exceptions import (
    ChisaError, ConversionError, DimensionMismatchError, AxiomViolationError, AmbiguousUnitError,
    UnitNotFoundError
)

from .core.registry import baseof, dims, unitsin, dimof
from .core import axioms as axiom
from .core import constants as const
from .core import vmath
from .core.axioms import C
from .core.engine import convert
from .core.base import BaseUnit

def _bootstrap_units() -> None:
    """
    Silently loads all built-in unit modules to trigger the Axiom Engine's
    auto-registration (__init_subclass__).
    """
    import importlib
    
    _core_dimensions = [
        "length", "mass", "pressure", "time", "speed", "temperature", 
        "data", "volume", "force", "energy", "power", "area"
    ]
    
    for dim in _core_dimensions:
        importlib.import_module(f".units.{dim}", package=__name__)


_bootstrap_units()
del _bootstrap_units


__version__ = "0.1.0"
__all__ = [
    "convert",
    "axiom",
    "C",
    "const",
    "vmath",
    "BaseUnit",
    "get_base_unit",
    "get_dimensions",
    "get_units_by_dimension",
    "dimension_of",
    "ChisaError",
    "ConversionError",
    "DimensionMismatchError",
    "AxiomViolationError",
    "AmbiguousUnitError",
    "UnitNotFoundError"
]


"""
Temperature and Thermodynamic Dimension Module.

This module defines scales for measuring thermal energy. 
It utilizes mathematical shift operations and absolute boundary constraints 
to prevent violations of physical laws (e.g., dropping below Absolute Zero).
While Kelvin is the absolute thermodynamic baseline, Celsius operates 
as the practical 1:1 scaling reference.
"""

from decimal import Decimal
from ..core.base import BaseUnit
from ..core import axioms as axiom
from ..core import constants as const

class TemperatureUnit(BaseUnit):
    """
    The primary parent class for the Temperature dimension.
    While Kelvin is the absolute thermodynamic baseline, Celsius operates 
    as the practical reference point with a 1.0 multiplier and 0.0 offset.
    """
    dimension = "temperature"


# =========================================================================
# 1. CELSIUS (BASE REFERENCE)
# =========================================================================
@axiom.bound(min_val=const.ABSOLUTE_ZERO_C, msg=f"Temperature cannot drop below Absolute Zero ({const.ABSOLUTE_ZERO_C} °C)!")
class Celsius(TemperatureUnit):
    symbol = "c"
    aliases = ["celsius", "celcius", "°c"]
    base_offset = 0.0
    base_multiplier = 1.0


# =========================================================================
# 2. FAHRENHEIT
# =========================================================================
@axiom.bound(min_val=const.ABSOLUTE_ZERO_F, msg=f"Physics Violation: Temperature cannot drop below Absolute Zero ({const.ABSOLUTE_ZERO_F} °F)!")
class Fahrenheit(TemperatureUnit):
    symbol = "f"
    aliases = ["fahrenheit", "°f"]
    # Offset: Minus freezing point
    base_offset = -const.FAHRENHEIT_OFFSET
    base_multiplier = Decimal('5') / Decimal('9')


# =========================================================================
# 3. KELVIN (ABSOLUTE THERMODYNAMIC SCALE)
# =========================================================================
@axiom.bound(min_val=const.ABSOLUTE_ZERO_K, msg=f"Physics Violation: Temperature cannot drop below Absolute Zero ({const.ABSOLUTE_ZERO_K} K)!")
class Kelvin(TemperatureUnit):
    symbol = "k"
    aliases = ["kelvin", "°k"]
    base_offset = -const.ZERO_CELSIUS_K
    base_multiplier = 1.0


# =========================================================================
# 4. RANKINE
# =========================================================================
@axiom.bound(min_val=const.ABSOLUTE_ZERO_K, msg=f"Physics Violation: Temperature cannot drop below Absolute Zero ({const.ABSOLUTE_ZERO_K} °R)!")
class Rankine(TemperatureUnit):
    symbol = "r"
    aliases = ["rankine", "°r"]
    base_offset = -const.RANKINE_OFFSET
    base_multiplier = Decimal('5') / Decimal('9')


# =========================================================================
# 5. REAUMUR
# =========================================================================
@axiom.bound(min_val=const.ABSOLUTE_ZERO_RE, msg=f"Physics Violation: Temperature cannot drop below Absolute Zero ({const.ABSOLUTE_ZERO_RE} °Re)!")
class Reaumur(TemperatureUnit):
    symbol = "re"
    aliases = ["reaumur", "réaumur", "°re", "°ré"]
    base_offset = 0.0
    # Multiplier: 5/4 = 1.25
    base_multiplier = 1.25


"""
Volume and 3D Space Dimension Module.

This module measures the three-dimensional space enclosed within a boundary.
It bridges derived geometric cubic lengths (e.g., Cubic Meter, Cubic Inch) 
with liquid capacity standards from metric, imperial, and US customary systems 
(e.g., Liters, Gallons, Barrels).
The absolute base unit for this dimension is the Cubic Meter (m³).
"""

from ..core.base import BaseUnit
from ..core import axioms as axiom
from ..core import constants as const
from .length import (
    Meter, Centimeter, Millimeter, Kilometer, Decimeter,
    Inch, Foot, Yard
)

@axiom.bound(min_val=0, msg="Volume (Space) cannot be negative!")
class VolumeUnit(BaseUnit):
    """
    The primary parent class for the Volume dimension.
    The base unit is strictly synthesized as Cubic Meter (m³).
    """
    dimension = "volume"


# =========================================================================
# 1. CUBIC LENGTHS
# =========================================================================
@axiom.derive(mul=[Meter, Meter, Meter])
class CubicMeter(VolumeUnit):
    symbol = "m³"
    aliases = ["m3", "cubicmeter", "cubic meters", "cubic meter", "cubicmeters"]

@axiom.derive(mul=[Decimeter, Decimeter, Decimeter])
class CubicDecimeter(VolumeUnit):
    symbol = "dm³"
    aliases = ["dm3", "cubic decimeter", "cubic decimeters"]

@axiom.derive(mul=[Centimeter, Centimeter, Centimeter])
class CubicCentimeter(VolumeUnit):
    symbol = "cm³"
    aliases = ["cm3", "cc", "cubic centimeter", "cubic centimeters", "cubic centimetre"]

@axiom.derive(mul=[Millimeter, Millimeter, Millimeter])
class CubicMillimeter(VolumeUnit):
    symbol = "mm³"
    aliases = ["mm3", "cubic millimeter", "cubic millimeters"]

@axiom.derive(mul=[Kilometer, Kilometer, Kilometer])
class CubicKilometer(VolumeUnit):
    symbol = "km³"
    aliases = ["km3", "cubic kilometer", "cubic kilometers"]

@axiom.derive(mul=[Inch, Inch, Inch])
class CubicInch(VolumeUnit):
    symbol = "in³"
    aliases = ["in3", "cubic inch", "cubic inches"]

@axiom.derive(mul=[Foot, Foot, Foot])
class CubicFoot(VolumeUnit):
    symbol = "ft³"
    aliases = ["ft3", "cubic foot", "cubic feet"]

@axiom.derive(mul=[Yard, Yard, Yard])
class CubicYard(VolumeUnit):
    symbol = "yd³"
    aliases = ["yd3", "cubic yard", "cubic yards"]


# =========================================================================
# 2. METRIC LIQUID VOLUME (Liters)
# =========================================================================
class Liter(VolumeUnit):
    symbol = "l"
    aliases = ["liter", "liters", "litre"]
    base_multiplier = const.LITER_TO_CUBIC_METER

class Hectoliter(VolumeUnit):
    symbol = "hl"
    aliases = ["hectoliter", "hectoliters"]
    base_multiplier = const.LITER_TO_CUBIC_METER * 100.0

class Dekaliter(VolumeUnit):
    symbol = "dal"
    aliases = ["dekaliter", "dekaliters"]
    base_multiplier = const.LITER_TO_CUBIC_METER * 10.0

class Deciliter(VolumeUnit):
    symbol = "dl"
    aliases = ["deciliter", "deciliters"]
    base_multiplier = const.LITER_TO_CUBIC_METER / 10.0

class Centiliter(VolumeUnit):
    symbol = "cl"
    aliases = ["centiliter", "centiliters"]
    base_multiplier = const.LITER_TO_CUBIC_METER / 100.0

class Milliliter(VolumeUnit):
    symbol = "ml"
    aliases = ["milliliter", "milliliters"]
    base_multiplier = const.LITER_TO_CUBIC_METER / 1000.0

class Microliter(VolumeUnit):
    symbol = "µl"
    aliases = ["ul", "microliter", "microliters"]
    base_multiplier = const.LITER_TO_CUBIC_METER / 1e6

class Nanoliter(VolumeUnit):
    symbol = "nl"
    aliases = ["nanoliter", "nanoliters"]
    base_multiplier = const.LITER_TO_CUBIC_METER / 1e9


# =========================================================================
# 3. US CUSTOMARY & IMPERIAL LIQUID VOLUME (NIST Standards)
# =========================================================================
class USGallon(VolumeUnit):
    symbol = "gal"
    aliases = ["us sgal", "gallon", "gallons", "us gal"]
    base_multiplier = const.US_GALLON_TO_CUBIC_METER

class USQuart(VolumeUnit):
    symbol = "qt"
    aliases = ["quart", "quarts"]
    # 1 Gallon = 4 Quarts
    base_multiplier = const.US_GALLON_TO_CUBIC_METER / 4.0

class USPint(VolumeUnit):
    symbol = "pt"
    aliases = ["pint", "pints"]
    # 1 Gallon = 8 Pints
    base_multiplier = const.US_GALLON_TO_CUBIC_METER / 8.0

class USCup(VolumeUnit):
    symbol = "cup"
    aliases = ["cups"]
    # 1 Gallon = 16 Cups
    base_multiplier = const.US_GALLON_TO_CUBIC_METER / 16.0

class USFluidOunce(VolumeUnit):
    symbol = "floz"
    aliases = ["fl oz", "fluid ounce", "fluid ounces"]
    # 1 Gallon = 128 Fluid Ounces
    base_multiplier = const.US_GALLON_TO_CUBIC_METER / 128.0

class USTablespoon(VolumeUnit):
    symbol = "tbsp"
    aliases = ["tablespoon", "tablespoons"]
    # 1 Fluid Ounce = 2 Tablespoons -> 1 Gallon = 256 Tablespoons
    base_multiplier = const.US_GALLON_TO_CUBIC_METER / 256.0

class USTeaspoon(VolumeUnit):
    symbol = "tsp"
    aliases = ["teaspoon", "teaspoons"]
    # 1 Tablespoon = 3 Teaspoons -> 1 Gallon = 768 Teaspoons
    base_multiplier = const.US_GALLON_TO_CUBIC_METER / 768.0

class UKGallon(VolumeUnit):
    symbol = "uk gal"
    aliases = ["uk-gal", "gal-uk", "imperial gallon", "imperial gallons"]
    base_multiplier = const.UK_GALLON_TO_CUBIC_METER

class OilBarrel(VolumeUnit):
    symbol = "bbl"
    aliases = ["barrel", "barrels", "oil barrel"]
    # 1 Standard Oil Barrel = 42 US Gallons
    base_multiplier = const.US_GALLON_TO_CUBIC_METER * 42.0


"""
Speed and Velocity Dimension Module.

This module measures the rate of change of position over time.
It supports linear velocities (m/s, mph), nautical speeds (Knots), 
cosmic constants (Speed of Light), and dynamically scaled environmental 
speeds (Mach, dependent on temperature).
The absolute base unit for this dimension is Meters per Second (m/s).
"""
from ..core.base import BaseUnit
from ..core import axioms as axiom
from ..core import constants as const
from ..core import vmath
from .temperature import Kelvin


@axiom.bound(min_val=0, msg="Speed (scalar magnitude) cannot be negative!")
class SpeedUnit(BaseUnit):
    """
    The primary parent class for the Speed dimension.
    The absolute base unit is Meter per Second (m/s).
    """
    dimension = "speed"

@axiom.prepare(temperature=Kelvin)
def _calc_mach_scale(temperature = const.STANDARD_ATMOSPHERE_TEMP_K) -> float:
    safe_temp = vmath.maximum(temperature, const.ABSOLUTE_ZERO_K)
    return const.SPEED_OF_SOUND_0C * vmath.sqrt(safe_temp / const.ZERO_CELSIUS_K)


# =========================================================================
# 1. METRIC UNITS (Base: Meters per second)
# =========================================================================
class MeterPerSecond(SpeedUnit):
    symbol = "m/s"
    aliases = ["meter per second", "meters per second"]
    base_multiplier = 1.0

class MeterPerMinute(SpeedUnit):
    symbol = "m/min"
    aliases = ["meter per minute", "meters per minute"]
    base_multiplier = 1.0 / 60.0

class KilometerPerHour(SpeedUnit):
    symbol = "km/h"
    aliases = ["kph", "kilometer per hour", "kilometers per hour", "km per hour"]
    base_multiplier = 1000.0 / 3600.0

class MillimeterPerSecond(SpeedUnit):
    symbol = "mm/s"
    aliases = ["millimeter per second", "millimeters per second", "mm per second"]
    base_multiplier = 0.001

class MillimeterPerMinute(SpeedUnit):
    symbol = "mm/min"
    aliases = ["millimeter per minute", "millimeters per minute", "mm per minute"]
    base_multiplier = 0.001 / 60.0

class CentimeterPerSecond(SpeedUnit):
    symbol = "cm/s"
    aliases = ["centimeter per second", "centimeters per second", "cm per second"]
    base_multiplier = 0.01

class CentimeterPerMinute(SpeedUnit):
    symbol = "cm/min"
    aliases = ["centimeter per minute", "centimeters per minute", "cm per minute"]
    base_multiplier = 0.01 / 60.0


# =========================================================================
# 2. IMPERIAL & US CUSTOMARY UNITS
# =========================================================================
class MilePerHour(SpeedUnit):
    symbol = "mi/h"
    aliases = ["mph", "mile per hour", "miles per hour"]
    # 1 mil/jam = (1 mil dalam meter) dibagi 3600 detik
    base_multiplier = const.MILE_TO_METER / 3600.0

class FootPerSecond(SpeedUnit):
    symbol = "ft/s"
    aliases = ["fps", "foot per second", "feet per second"]
    base_multiplier = const.FOOT_TO_METER

class FootPerMinute(SpeedUnit):
    symbol = "ft/min"
    aliases = ["fpm", "foot per minute", "feet per minute"]
    base_multiplier = const.FOOT_TO_METER / 60.0

class InchPerSecond(SpeedUnit):
    symbol = "in/s"
    aliases = ["inch per second", "inches per second"]
    base_multiplier = const.INCH_TO_METER


# =========================================================================
# 3. NAUTICAL & AEROSPACE UNITS
# =========================================================================
class Knot(SpeedUnit):
    symbol = "kt"
    aliases = ["knot", "knots", "kn"]
    base_multiplier = const.NAUTICAL_MILE_TO_METER / 3600.0

@axiom.scale(formula=_calc_mach_scale)
class Mach(SpeedUnit):
    symbol = "mach"
    aliases = ["ma"]
    base_multiplier = 1.0


# =========================================================================
# 4. COSMIC / THEORETICAL UNITS
# =========================================================================
class SpeedOfLight(SpeedUnit):
    symbol = "c"
    aliases = ["speed of light", "lightspeed"]
    base_multiplier = const.SPEED_OF_LIGHT



"""
Length and Distance Dimension Module.

This module defines units for measuring one-dimensional physical space. 
It spans an extreme range of magnitudes, from microscopic and atomic scales 
(e.g., Nanometer, Angstrom) to human-scale imperial/metric systems, up to 
astronomical distances (e.g., Light Year, Parsec).
The absolute base unit for this dimension is the Meter (m).
"""

from ..core.base import BaseUnit
from ..core import axioms as axiom
from ..core import constants as const

@axiom.bound(min_val=0, msg="Length cannot be a negative value.")
class LengthUnit(BaseUnit):
    """
    The primary parent class for the Length dimension.
    The base unit is Meter (m).
    """
    dimension = "length"


# =========================================================================
# 1. METRIC UNITS (SI)
# (Metric is intrinsically base-10, so magic numbers are acceptable here)
# =========================================================================
class Meter(LengthUnit):
    symbol = "m"
    aliases = ["meter", "meters"]
    base_multiplier = 1.0

class Kilometer(LengthUnit):
    symbol = "km"
    aliases = ["kilometer", "kilometers"]
    base_multiplier = 1000.0

class Decimeter(LengthUnit):
    symbol = "dm"
    aliases = ["decimeter", "decimeters"]
    base_multiplier = 1e-1

class Centimeter(LengthUnit):
    symbol = "cm"
    aliases = ["centimeter", "centimeters"]
    base_multiplier = 1e-2

class Millimeter(LengthUnit):
    symbol = "mm"
    aliases = ["millimeter", "millimeters"]
    base_multiplier = 1e-3

class Micrometer(LengthUnit):
    symbol = "um"
    aliases = ["µm", "micrometer", "micrometers"]
    base_multiplier = 1e-6

class Nanometer(LengthUnit):
    symbol = "nm"
    aliases = ["nanometer", "nanometers"]
    base_multiplier = 1e-9

class Picometer(LengthUnit):
    symbol = "pm"
    aliases = ["picometer", "picometers"]
    base_multiplier = 1e-12

class Femtometer(LengthUnit):
    symbol = "fm"
    aliases = ["femtometer", "fermi"]
    base_multiplier = 1e-15

class Angstrom(LengthUnit):
    symbol = "Å"
    aliases = ["angstrom"]
    base_multiplier = 1e-10


# =========================================================================
# 2. ASTRONOMICAL UNITS
# =========================================================================
class LightYear(LengthUnit):
    symbol = "ly"
    aliases = ["lightyear", "lightyears"]
    base_multiplier = const.LIGHT_YEAR_METER

class AstronomicalUnit(LengthUnit):
    symbol = "au"
    aliases = ["astronomicalunit", "astronomicalunits"]
    base_multiplier = const.ASTRONOMICAL_UNIT_METER

class Parsec(LengthUnit):
    symbol = "pc"
    aliases = ["parsec", "parsecs"]
    base_multiplier = const.PARSEC_METER


# =========================================================================
# 3. IMPERIAL / US UNITS
# =========================================================================
class Inch(LengthUnit):
    symbol = "in"
    aliases = ["inch", "inches"]
    base_multiplier = const.INCH_TO_METER

class Foot(LengthUnit):
    symbol = "ft"
    aliases = ["foot", "feet"]
    base_multiplier = const.FOOT_TO_METER

class Yard(LengthUnit):
    symbol = "yd"
    aliases = ["yard", "yards"]
    # 1 Yard = 3 Feet
    base_multiplier = const.FOOT_TO_METER * 3.0

class Mile(LengthUnit):
    symbol = "mi"
    aliases = ["mile", "miles"]
    base_multiplier = const.MILE_TO_METER

class NauticalMile(LengthUnit):
    symbol = "nmi"
    aliases = ["nauticalmile", "nauticalmiles"]
    base_multiplier = const.NAUTICAL_MILE_TO_METER

class Mil(LengthUnit):
    symbol = "mil"
    aliases = ["thou"]
    # 1 Mil = 1/1000 Inch
    base_multiplier = const.INCH_TO_METER / 1000.0

class League(LengthUnit):
    symbol = "league"
    aliases = ["leagues"]
    # 1 League = 3 Miles
    base_multiplier = const.MILE_TO_METER * 3.0

class Hand(LengthUnit):
    symbol = "hand"
    aliases = ["hands"]
    # 1 Hand = 4 Inches
    base_multiplier = const.INCH_TO_METER * 4.0

class Barleycorn(LengthUnit):
    symbol = "barleycorn"
    aliases = ["barleycorns"]
    # 1 Barleycorn = 1/3 Inch
    base_multiplier = const.INCH_TO_METER / 3.0


# =========================================================================
# 4. ENGINEERING / SURVEY UNITS
# =========================================================================
class Chain(LengthUnit):
    symbol = "chain"
    aliases = ["chains"]
    # 1 Chain = 66 Feet
    base_multiplier = const.FOOT_TO_METER * 66.0

class Link(LengthUnit):
    symbol = "link"
    aliases = ["links"]
    # 1 Link = 0.66 Feet
    base_multiplier = const.FOOT_TO_METER * 0.66

class Rod(LengthUnit):
    symbol = "rod"
    aliases = ["rods", "pole", "poles", "perch", "perches"]
    # 1 Rod = 16.5 Feet
    base_multiplier = const.FOOT_TO_METER * 16.5

class Furlong(LengthUnit):
    symbol = "furlong"
    aliases = ["furlongs"]
    # 1 Furlong = 660 Feet
    base_multiplier = const.FOOT_TO_METER * 660.0


# =========================================================================
# 5. TYPOGRAPHIC UNITS
# =========================================================================
class Point(LengthUnit):
    symbol = "pt"
    aliases = ["point", "points"]
    # DTP Point = 1/72 Inch
    base_multiplier = const.INCH_TO_METER / 72.0

class Pica(LengthUnit):
    symbol = "pica"
    aliases = ["picas"]
    # 1 Pica = 1/6 Inch (or 12 Points)
    base_multiplier = const.INCH_TO_METER / 6.0


from decimal import Decimal, getcontext
from typing import Union, Optional, List, Any, Type, Dict
from .registry import default_ureg
from ..exceptions import DimensionMismatchError
from .axioms import _normalize_types

try:
    import numpy as np
    HAS_NUMPY = True
except ImportError:
    HAS_NUMPY = False


class BaseUnit:
    """
    The foundational core class for all physical and digital units in Chisa.
    Provides the standard interface for dimensional algebra, scalar conversions,
    NumPy vectorization, and fluent formatting.
    """
    dimension: Optional[str] = None
    symbol: Optional[str] = None
    aliases: Optional[List[str]] = None
    base_multiplier: Union[float, Decimal] = 1.0
    base_offset: Union[float, Decimal] = 0.0

    def __init_subclass__(cls, **kwargs) -> None:
        """
        Automatically triggered when a subclass of BaseUnit is created.
        Injects the new class directly into the global UnitRegistry.
        """
        super().__init_subclass__(**kwargs)
        default_ureg._register(cls)

    def __init__(self, value: Union[int, float, str, Decimal, Any], context: Optional[Dict[str, Any]] = None) -> None:
        """
        A unit object, handling both scalar and vectorized inputs.

        Args:
            value (Union[int, float, str, Decimal, Any]): The magnitude of the unit. 
                Accepts pure scalars, lists, tuples, or NumPy arrays.
            context (Optional[Dict[str, Any]]): Contextual physical properties 
                required for dynamic scaling (e.g., {'temp_c': 30}).
                
        Raises:
            TypeError: If the input value is not a recognized numeric or iterable type.
        """
        if HAS_NUMPY and isinstance(value, (np.ndarray, np.generic)):
            self._value = value
        elif HAS_NUMPY and isinstance(value, (list, tuple)):
            self._value = np.array(value, dtype=float)
        elif isinstance(value, (int, float, str, Decimal)):
            self._value = Decimal(str(value))
        else:
            raise TypeError(f"Value must be numeric or array, got {type(value).__name__}")
            
        self.context = context or {}

    # =========================================================================
    # PUBLIC PROPERTIES (THE DX INTERFACE)
    # =========================================================================
    @property
    def exact(self) -> Union[Decimal, Any]:
        """
        Returns the exact, unadulterated internal representation of the unit.
        
        - For scalars: Returns a `decimal.Decimal` to preserve absolute mathematical 
          precision (avoiding floating-point drift).
        - For vectors: Returns the original `numpy.ndarray`.
        
        WARNING: Do not mix `.exact` scalar outputs with external NumPy arrays 
        in manual calculations, as Python forbids multiplying Decimals with floats/arrays. 
        Use `.magnitude` instead for general cross-dimensional math.
        """
        return self._value

    @property
    def mag(self) -> Any:
        """
        Returns the numeric magnitude as a standard, math-safe Python primitive.
        
        - For scalars: Automatically casts internal Decimals down to standard `float` or `int`.
        - For vectors: Returns the `numpy.ndarray` intact.
        
        Always use `.magnitude` when extracting values to perform 
        external physics calculations, plotting (Matplotlib), or machine learning.
        """
        if HAS_NUMPY and isinstance(self._value, (np.ndarray, np.generic)):
            return self._value
            
        if isinstance(self._value, Decimal) and self._value == int(self._value):
            return int(self._value)
            
        return float(self._value)

    # =========================================================================
    # INTERNAL CORE LOGIC
    # =========================================================================
    def _to_base_value(self) -> Union[Decimal, Any]:
        """Converts the current magnitude to the dimension's absolute base point."""
        if HAS_NUMPY and isinstance(self._value, (np.ndarray, np.generic)):
            val_with_offset = self._value + float(self.base_offset)
            return val_with_offset * float(self.base_multiplier)
        
        val_with_offset = self._value + Decimal(str(self.base_offset))
        return val_with_offset * Decimal(str(self.base_multiplier))

    @classmethod
    def _from_base_value(cls, base_val: Union[Decimal, Any], context: Dict[str, Any]) -> Union[Decimal, Any]:
        """Converts a magnitude from the dimension's absolute base point to this specific unit."""
        if HAS_NUMPY and isinstance(base_val, (np.ndarray, np.generic, float)):
            val = base_val / float(cls.base_multiplier)
            return val - float(cls.base_offset)
            
        val = base_val / Decimal(str(cls.base_multiplier))
        return val - Decimal(str(cls.base_offset))

    # Whitelisted NumPy Operations
    _ALLOWED_UFUNCS = {
        'add', 'subtract', 'maximum', 'minimum', 'fmax', 'fmin', 
        'absolute', 'negative', 'positive', 'rint'
    }
    
    _ALLOWED_FUNCTIONS = {
        'sum', 'max', 'min', 'mean', 'median', 'std', 'round', 'ptp'
    }

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Portal for low-level element-wise operations (e.g., np.add)."""
        if ufunc.__name__ not in self._ALLOWED_UFUNCS:
            raise TypeError(
                f"Chisa Dimensional Guard: Operation 'np.{ufunc.__name__}' is "
                f"blocked because it mutates physical dimensions."
            )

        raw_inputs = []
        for inp in inputs:
            if isinstance(inp, BaseUnit):
                if self.dimension != inp.dimension:
                    raise DimensionMismatchError(str(self.dimension), str(inp.dimension), f"NumPy ufunc {ufunc.__name__}")
                raw_inputs.append(inp._to_base_value())
            else:
                raw_inputs.append(inp)

        raw_result = getattr(ufunc, method)(*raw_inputs, **kwargs)
        final_value = self.__class__._from_base_value(raw_result, self.context)
        return self.__class__(final_value, context=self.context)
    
    def __array_function__(self, func, types, args, kwargs):
        """Portal for high-level reduction/statistical operations (e.g., np.mean, np.max)."""
        if func.__name__ not in self._ALLOWED_FUNCTIONS:
            raise TypeError(
                f"Chisa Dimensional Guard: Function 'np.{func.__name__}' is "
                f"blocked to preserve dimensional integrity."
            )

        def _extract_raw(obj):
            if isinstance(obj, BaseUnit):
                return obj._to_base_value()
            if isinstance(obj, (list, tuple)):
                return [_extract_raw(o) for o in obj]
            return obj

        raw_args = tuple(_extract_raw(arg) for arg in args)
        raw_result = func(*raw_args, **kwargs)

        final_value = self.__class__._from_base_value(raw_result, self.context)
        return self.__class__(final_value, context=self.context)
    
    def __str__(self) -> str:
        if HAS_NUMPY and isinstance(self._value, np.ndarray):
            return f"<{self.__class__.__name__} Array: {self.format(tag=False)}>"
        elif (HAS_NUMPY and isinstance(self._value, np.generic)) or isinstance(self._value, float):
            return self.format(tag=True)
            
        return self.format(tag=True)

    def __repr__(self) -> str:
        if HAS_NUMPY and isinstance(self._value, np.ndarray):
            return f"<{self.__class__.__name__} Array of shape {self._value.shape}>"
        return f"<{self.__class__.__name__}: {self._value} {self.symbol}>"

    def __float__(self) -> float:
        return float(self._value)

    def __int__(self) -> int:
        return int(self._value)

    def to(self, unit: Union[Type['BaseUnit'], str]) -> 'BaseUnit':
        """
        Converts the current OOP unit object into another unit within the same dimension.
        Elegantly resolves string aliases via the global registry.

        Args:
            unit: The target destination class or string alias.

        Returns:
            A new instance of the target unit class maintaining the original context.
            
        Raises:
            TypeError: If the target is not a valid Chisa BaseUnit or recognized string.
            DimensionMismatchError: If the target belongs to a different physical dimension.
        """
        if isinstance(unit, str):
            try:
                from .registry import default_ureg 
                TargetClass = default_ureg.get_unit_class(unit, expected_dim=self.dimension)
            except Exception as e:
                raise TypeError(
                    f"Invalid target in .to() method: '{unit}'. "
                    f"Please provide a valid BaseUnit class or a recognized string alias for dimension '{self.dimension}'."
                ) from None
        else:
            TargetClass = unit

        if not (isinstance(TargetClass, type) and issubclass(TargetClass, BaseUnit)):
            raise TypeError(
                f"The .to() method expects a BaseUnit class or a valid string alias, "
                f"but got {type(unit).__name__} instead."
            )
        if self.dimension != getattr(TargetClass, "dimension", None):
            raise DimensionMismatchError(str(self.dimension), str(getattr(TargetClass, 'dimension', None)), "Conversion")
            
        base_val = self._to_base_value()
        new_val = TargetClass._from_base_value(base_val, self.context)
        return TargetClass(new_val, context=self.context)

    def format(
        self,
        prec: int = 4,
        sigfigs: Optional[int] = None,
        scinote: bool = False,
        delim: Union[bool, str] = False,
        tag: bool = True
    ) -> str:
        """
        Formats the object's scalar value into a structured, human-readable string.
        Safely applies string manipulation for vectorized NumPy arrays.

        Args:
            prec: Number of decimal places to display. Defaults to 4.
            sigfigs: Number of significant figures to enforce.
            scinote: If True, forces scientific notation (e.g., 1.5E3).
            delim: If True, applies thousands separators. Can be a custom string character.
            tag: If True, appends the unit's symbol to the end of the string.

        Returns:
            str: The formatted cosmetic string representing the unit.
        """
        val = self._value
        
        if HAS_NUMPY and isinstance(val, (np.ndarray, np.generic)):
            def format_elem(x):
                if scinote:
                    digits = sigfigs - 1 if sigfigs is not None else prec
                    return f"{x:.{digits}E}"
                
                s = f"{float(x):.{prec}f}"
                if '.' in s:
                    s = s.rstrip('0').rstrip('.')
                    if s.endswith('-'): s = s + "0"
                    if not s: s = "0"
                
                if delim:
                    separator = "," if delim is True or str(delim).lower() == "default" else str(delim)
                    parts = s.split('.')
                    parts[0] = f"{int(parts[0]):,}".replace(",", separator)
                    s = '.'.join(parts) if len(parts) > 1 else parts[0]
                    
                return s

            val_str = np.array2string(
                val, 
                formatter={'float_kind': format_elem, 'int': format_elem},
                separator=', ',
                suppress_small=not scinote 
            )
            
            if tag:
                return f"{val_str} {self.symbol}"
            return val_str

        if sigfigs is not None:
            if val == 0:
                val = Decimal(0)
            else:
                shift = sigfigs - val.adjusted() - 1
                quantizer = Decimal('1e{}'.format(-shift))
                val = val.quantize(quantizer, rounding=getcontext().rounding)

        if scinote:
            digits = sigfigs - 1 if sigfigs is not None else prec
            val_str = f"{val:.{digits}E}"
        else:
            if sigfigs:
                decimal_places = max(sigfigs - (val.adjusted() + 1), 0)
                val_str = f"{val:.{decimal_places}f}"
            else:
                try:
                    quant = Decimal(f"1e-{prec}")
                    val = val.quantize(quant, rounding=getcontext().rounding)
                except Exception:
                    pass
                val_str = format(val, 'f')

        if not scinote and '.' in val_str:
            val_str = val_str.rstrip('0').rstrip('.')

        if delim:
            separator = "," if delim is True or str(delim).lower() == "default" else str(delim)
            if not scinote:
                parts = val_str.split('.')
                parts[0] = f"{int(parts[0]):,}".replace(",", separator)
                val_str = '.'.join(parts) if len(parts) > 1 else parts[0]

        if tag:
            return f"{val_str} {self.symbol}"
        return val_str

    def __eq__(self, other: Any) -> Union[bool, Any]:
        if not isinstance(other, BaseUnit):
            return NotImplemented
        if self.dimension != getattr(other, "dimension", None):
            raise DimensionMismatchError(str(self.dimension), str(getattr(other, 'dimension', None)), "Equality (==)")
            
        v1, v2 = self._to_base_value(), other._to_base_value()
        v1, v2 = _normalize_types(v1, v2)
            
        return v1 == v2

    def __lt__(self, other: Any) -> Union[bool, Any]:
        if not isinstance(other, BaseUnit):
            return NotImplemented
        if self.dimension != getattr(other, "dimension", None):
            raise DimensionMismatchError(str(self.dimension), str(getattr(other, 'dimension', None)), "Less-than (<)")
            
        v1, v2 = self._to_base_value(), other._to_base_value()
        v1, v2 = _normalize_types(v1, v2)
            
        return v1 < v2

    def __le__(self, other: Any) -> Union[bool, Any]:
        if not isinstance(other, BaseUnit):
            return NotImplemented
        if self.dimension != getattr(other, "dimension", None):
            raise DimensionMismatchError(str(self.dimension), str(getattr(other, 'dimension', None)), "Less-than-or-equal (<=)")
            
        v1, v2 = self._to_base_value(), other._to_base_value()
        v1, v2 = _normalize_types(v1, v2)
            
        return v1 <= v2

    def __ge__(self, other: Any) -> Union[bool, Any]:
        if not isinstance(other, BaseUnit):
            return NotImplemented
        if self.dimension != getattr(other, "dimension", None):
            raise DimensionMismatchError(str(self.dimension), str(getattr(other, 'dimension', None)), "Greater-than-or-equal (>=)")
            
        v1, v2 = self._to_base_value(), other._to_base_value()
        v1, v2 = _normalize_types(v1, v2)
            
        return v1 >= v2

    def __ne__(self, other: Any) -> Union[bool, Any]:
        if not isinstance(other, BaseUnit):
            return NotImplemented
        if self.dimension != getattr(other, "dimension", None):
            raise DimensionMismatchError(str(self.dimension), str(getattr(other, 'dimension', None)), "Not-Equal (!=)")
            
        v1, v2 = self._to_base_value(), other._to_base_value()
        v1, v2 = _normalize_types(v1, v2)
            
        return v1 != v2
    
    def __add__(self, other: Any) -> 'BaseUnit':
        if not isinstance(other, BaseUnit):
            return NotImplemented
        if self.dimension != getattr(other, "dimension", None):
            raise DimensionMismatchError(str(self.dimension), str(getattr(other, 'dimension', None)), "Addition (+)")
        
        v1, v2 = self._to_base_value(), other._to_base_value()
        v1, v2 = _normalize_types(v1, v2)
        
        total_base = v1 + v2
        merged_context = {**other.context, **self.context}
        final_value = self.__class__._from_base_value(total_base, merged_context)
        return self.__class__(final_value, context=merged_context)

    def __sub__(self, other: Any) -> 'BaseUnit':
        if not isinstance(other, BaseUnit):
            return NotImplemented
        if self.dimension != getattr(other, "dimension", None):
            raise DimensionMismatchError(str(self.dimension), str(getattr(other, 'dimension', None)), "Subtraction (-)")
        
        v1, v2 = self._to_base_value(), other._to_base_value()
        v1, v2 = _normalize_types(v1, v2)
            
        total_base = v1 - v2
        merged_context = {**other.context, **self.context}
        final_value = self.__class__._from_base_value(total_base, merged_context)
        return self.__class__(final_value, context=merged_context)
    
    def __mul__(self, other: Any) -> 'BaseUnit':
        """Permits multiplication of a unit strictly by a pure scalar or array."""
        if isinstance(other, (int, float, Decimal)) or (HAS_NUMPY and isinstance(other, (np.ndarray, np.generic))):
            new_val = self._value * other
            return self.__class__(new_val, context=self.context)
        
        if isinstance(other, BaseUnit):
            raise NotImplementedError(
                "Cross-dimensional multiplication (e.g., Meter * Meter) is restricted "
                "in this version to preserve dimensional integrity."
            )
        return NotImplemented

    def __rmul__(self, other: Any) -> 'BaseUnit':
        """Supports reverse scalar multiplication (e.g., 3 * Meter(5))."""
        return self.__mul__(other)

    def __truediv__(self, other: Any) -> 'BaseUnit':
        """Permits division of a unit strictly by a pure scalar or array."""
        if isinstance(other, (int, float, Decimal)) or (HAS_NUMPY and isinstance(other, (np.ndarray, np.generic))):
            new_val = self._value / other
            return self.__class__(new_val, context=self.context)
            
        if isinstance(other, BaseUnit):
            raise NotImplementedError(
                "Cross-dimensional division (e.g., Meter / Second) is restricted "
                "in this version to preserve dimensional integrity."
            )
        return NotImplemented

    def __array__(self, dtype=None) -> Any:
        """
        NumPy standard dunder method to expose the raw array.
        Allows Chisa objects to be natively ingested by Matplotlib, Scikit-Learn, pandas, or SymPy.
        """
        if HAS_NUMPY:
            return np.asarray(self._value, dtype=dtype)
        return np.array([self._value], dtype=dtype)